# Generated from BOCL.g4 by ANTLR 4.13.1
from antlr4 import *

if "." in __name__:
    from .BOCLParser import BOCLParser
else:
    from BOCLParser import BOCLParser
import inspect
from besser.BUML.metamodel.ocl.ocl import OperationCallExpression
from besser.BUML.notations.ocl.comparison_operator_checker import comparison_verifier

# This class defines a complete listener for a parse tree produced by BOCLParser.
class BOCLListener(ParseTreeListener):
    operator = None

    def __init__(self, rh):
        self.rootHandler = rh
        self.forAllBody = False
        self.operator = []
        self.initItems = []
        self.coll_data = []
        self.primaryExp =None
        self.debug = False
        self.debug_print = False
        self.all_if_else = []
        self.types_of = []
        self.userDefined = None
        self.unary = ""
        self.sign = []
        self.num = []
        self.functions = []
    def preprocess(self,ocl):
        if ocl.count('(') != ocl.count(')'):
            raise Exception(" Incorrect Syntax: Number of Brackets Mismatch")
        if ocl.count('{') != ocl.count('}'):
            raise Exception(" Incorrect Syntax: Number of Brackets Mismatch")
        if ocl.count('[') != ocl.count(']'):
            raise Exception(" Incorrect Syntax: Number of Brackets Mismatch")
        cc = comparison_verifier()

        # print(cc.verify_ocl(ocl))
        if not cc.verify_ocl(ocl):
            raise Exception("Incorrect Syntax for comparison operators")
    # Enter a parse tree produced by BOCLParser#oclFile.
    def enterOclFile(self, ctx: BOCLParser.OclFileContext):
        # print(inspect.stack()[0][3])
        self.preprocess(ctx.getText())
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        if 'inv' in ctx.getText().split(':')[0]:
            self.rootHandler.set_inv(True)
        if 'pre:' in ctx.getText():
            self.rootHandler.set_pre(True)
        if 'post:' in ctx.getText():
            self.rootHandler.set_post(True)
        if 'body:' in ctx.getText():
            self.rootHandler.set_post(True)

        pass

    # Exit a parse tree produced by BOCLParser#oclFile.
    def exitOclFile(self, ctx: BOCLParser.OclFileContext):
        if len(self.operator) == 1:
            if self.userDefined is not None:
                self.rootHandler.handle_last_opnum(self.operator.pop(),self.userDefined)
            else:
                pass #throw exception
        else:
            pass  # throw exception


        pass

        # Enter a parse tree produced by BOCLParser#preCondition.
    def enterPreCondition(self, ctx: BOCLParser.PreConditionContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        # Exit a parse tree produced by BOCLParser#preCondition.
    def exitPreCondition(self, ctx: BOCLParser.PreConditionContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        context = ctx.getText().split("context")[1].split(":")[0]
        function_name = ctx.getText().split("::")[1].split("pre:")[0]
        self.rootHandler.handle_pre_root(context, function_name, self.rootHandler.get_root())

        # Enter a parse tree produced by BOCLParser#postCondition.
    def enterPostCondition(self, ctx: BOCLParser.PostConditionContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        # Exit a parse tree produced by BOCLParser#postCondition.
    def exitPostCondition(self, ctx: BOCLParser.PostConditionContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        context = ctx.getText().split("context")[1].split(":")[0]
        function_name = ctx.getText().split("::")[1].split("post:")[0]
        self.rootHandler.handle_post_root(context, function_name, self.rootHandler.get_root())

        # Enter a parse tree produced by BOCLParser#initConstraints.
    def enterInitConstraints(self, ctx: BOCLParser.InitConstraintsContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        # Exit a parse tree produced by BOCLParser#initConstraints.
    def exitInitConstraints(self, ctx: BOCLParser.InitConstraintsContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())

        context = ctx.getText().split("context")[1].split(":")[0]
        variable_name = ctx.getText().split("::")[1].split(":")[0]
        variable_type = ctx.getText().split("init:")[0].split(":")[-1]
        # print(ctx.getText())
        if len(self.initItems)>0 and self.rootHandler.get_root() is None:
            self.rootHandler.handle_init_root(context, variable_name, self.initItems.pop(-1), variable_type)
        else:
            self.rootHandler.handle_init_root(context, variable_name, self.rootHandler.get_root(),variable_type)
    # Enter a parse tree produced by BOCLParser#ContextExp.
    def enterContextExp(self, ctx: BOCLParser.ContextExpContext):
        self.context = (ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#ContextExp.
    def exitContextExp(self, ctx: BOCLParser.ContextExpContext):
        # print(inspect.stack()[0][3])

        pass

    # Enter a parse tree produced by BOCLParser#constraint.
    def enterConstraint(self, ctx: BOCLParser.ConstraintContext):

        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        self.context_name = (self.context.split(ctx.getText()))[0].replace('context', '')
        self.rootHandler.set_context_name(self.context_name)
        # print(ctx.getText())

    # Exit a parse tree produced by BOCLParser#constraint.
    def exitConstraint(self, ctx: BOCLParser.ConstraintContext):
        pass

    # Enter a parse tree produced by BOCLParser#functionCall.
    def enterFunctionCall(self, ctx: BOCLParser.FunctionCallContext):
        if self.debug:
            print(inspect.stack()[0][3])
        if self.debug_print:
            print(ctx.getText())
        op_exp = self.rootHandler.create_operation_call_exp(ctx.getText())
        op_exp.referredOperation=self.rootHandler.create_operation_call_exp('FunctionCall')
        self.types_of.append(op_exp)
        self.functions.append(op_exp)
        # print(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#functionCall.
    def exitFunctionCall(self, ctx: BOCLParser.FunctionCallContext):
        # print(inspect.stack()[0][3])
        pass

    # Enter a parse tree produced by BOCLParser#type.
    def enterType(self, ctx: BOCLParser.TypeContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass
        type_exp = self.rootHandler.create_type_exp(ctx.getText())
        self.types_of[-1].add(type_exp)
        if isinstance(self.types_of[-1].referredOperation, OperationCallExpression):
            if self.types_of[-1].referredOperation.name == "FunctionCall":
                self.types_of.pop()

    # Exit a parse tree produced by BOCLParser#type.
    def exitType(self, ctx: BOCLParser.TypeContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Enter a parse tree produced by BOCLParser#collectionType.
    def enterCollectionType(self, ctx: BOCLParser.CollectionTypeContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#collectionType.
    def exitCollectionType(self, ctx: BOCLParser.CollectionTypeContext):
        pass

    # Enter a parse tree produced by BOCLParser#userDefinedType.
    def enterUserDefinedType(self, ctx: BOCLParser.UserDefinedTypeContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#userDefinedType.
    def exitUserDefinedType(self, ctx: BOCLParser.UserDefinedTypeContext):
        pass

    # Enter a parse tree produced by BOCLParser#binary.
    def enterBinary(self, ctx: BOCLParser.BinaryContext):
        # print(inspect.stack()[0][3])
        self.binary = ctx.getText()

        if self.unary:
            self.sign.append(self.unary.split(self.binary)[0])
            # print(self.sign)
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        # print(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#binary.
    def exitBinary(self, ctx: BOCLParser.BinaryContext):

        pass

    # Enter a parse tree produced by BOCLParser#unary.
    def enterUnary(self, ctx: BOCLParser.UnaryContext):
        # print(inspect.stack()[0][3])4
        self.unary = ctx.getText()
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#unary.
    def exitUnary(self, ctx: BOCLParser.UnaryContext):
        pass


    # Enter a parse tree produced by BOCLParser#OCLISTYPEOF.
    def enterOCLISTYPEOF(self, ctx: BOCLParser.OCLISTYPEOFContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        if ctx.parentCtx is not None:

            propertyName = ctx.parentCtx.getText().split('.')[-2]
            self.rootHandler.handle_property(propertyName)
        op_exp = self.rootHandler.create_operation_call_exp("OCLISTYPEOF")
        op_exp.referredOperation=self.rootHandler.create_operation_call_exp('OCLISTYPEOF')
        self.types_of.append(op_exp)
        pass

    # Exit a parse tree produced by BOCLParser#OCLISTYPEOF.
    def exitOCLISTYPEOF(self, ctx: BOCLParser.OCLISTYPEOFContext):
        self.rootHandler.add_to_root(self.types_of.pop())
        pass

    # Enter a parse tree produced by BOCLParser#OCLASTYPE.
    def enterOCLASTYPE(self, ctx: BOCLParser.OCLASTYPEContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_exp = self.rootHandler.create_operation_call_exp("OCLASTYPE")
        op_exp.referredOperation=self.rootHandler.create_operation_call_exp('OCLASTYPE')
        self.types_of.append(op_exp)

        pass

    # Exit a parse tree produced by BOCLParser#OCLASTYPE.
    def exitOCLASTYPE(self, ctx: BOCLParser.OCLASTYPEContext):
        self.rootHandler.add_to_root(self.types_of.pop())
        pass

    # Enter a parse tree produced by BOCLParser#OCLISKINDOF.
    def enterOCLISKINDOF(self, ctx: BOCLParser.OCLISKINDOFContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_exp = self.rootHandler.create_operation_call_exp("OCLISKINDOF")
        op_exp.referredOperation=self.rootHandler.create_operation_call_exp('OCLISKINDOF')
        self.types_of.append(op_exp)

        pass

    # Exit a parse tree produced by BOCLParser#OCLISKINDOF.
    def exitOCLISKINDOF(self, ctx: BOCLParser.OCLISKINDOFContext):
        self.rootHandler.add_to_root(self.types_of.pop())
        pass

    # Enter a parse tree produced by BOCLParser#ISEMPTY.
    def enterISEMPTY(self, ctx: BOCLParser.ISEMPTYContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('IsEmpty')
        if len(self.coll_data) != 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        else:
            if ctx.parentCtx is not None:
                propertyName = ctx.parentCtx.getText().split('->')[0]
                self.rootHandler.handle_property(propertyName)
        self.rootHandler.handle_adding_to_root(op_call_exp)

        pass

    # Exit a parse tree produced by BOCLParser#ISEMPTY.
    def exitISEMPTY(self, ctx: BOCLParser.ISEMPTYContext):
        pass

    # Enter a parse tree produced by BOCLParser#SUM.
    def enterSUM(self, ctx: BOCLParser.SUMContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('Sum')
        if len(self.coll_data) != 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        else:
            if ctx.parentCtx is not None:
                propertyName = ctx.parentCtx.getText().split('->')[0]
                self.rootHandler.handle_property(propertyName)

        self.rootHandler.handle_adding_to_root(op_call_exp)

        pass

    # Exit a parse tree produced by BOCLParser#SUM.
    def exitSUM(self, ctx: BOCLParser.SUMContext):
        pass

    # Enter a parse tree produced by BOCLParser#SIZE.
    def enterSIZE(self, ctx: BOCLParser.SIZEContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('Size')
        if len(self.coll_data) != 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        else:
            if ctx.parentCtx is not None:
                propertyName = ctx.parentCtx.getText().split('->')[0]
                if len(propertyName)>1:
                    self.rootHandler.handle_property(propertyName)
            # self.rootHandler.handle_property(propertyName)
        self.rootHandler.handle_adding_to_root(op_call_exp)

        pass

    # Exit a parse tree produced by BOCLParser#SIZE.
    def exitSIZE(self, ctx: BOCLParser.SIZEContext):
        pass

    # Enter a parse tree produced by BOCLParser#INCLUDES.
    def enterINCLUDES(self, ctx: BOCLParser.INCLUDESContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('INCLUDES')
        if len(self.coll_data) != 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        else:
            if ctx.parentCtx is not None:
                propertyName = ctx.parentCtx.getText().split('->')[0]
                self.rootHandler.handle_property(propertyName)
        self.coll_data.append(op_call_exp)

        pass

    # Exit a parse tree produced by BOCLParser#INCLUDES.
    def exitINCLUDES(self, ctx: BOCLParser.INCLUDESContext):

        exp = self.coll_data.pop()
        if self.primaryExp is not None:
            if len(exp.arguments) == 0:
                prop = self.rootHandler.create_property(self.primaryExp)
                exp.arguments.append(prop)
        self.rootHandler.add_to_root(exp)
        pass

    # Enter a parse tree produced by BOCLParser#EXCLUDES.
    def enterEXCLUDES(self, ctx: BOCLParser.EXCLUDESContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('EXCLUDES')
        if len(self.coll_data) != 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        else:
            temp = ctx.getText()
            if ctx.parentCtx is not None:
                propertyName = ctx.parentCtx.getText().split('->')[0]
                self.rootHandler.handle_property(propertyName)
        self.coll_data.append(op_call_exp)

        pass

    # Exit a parse tree produced by BOCLParser#EXCLUDES.
    def exitEXCLUDES(self, ctx: BOCLParser.EXCLUDESContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        exp = self.coll_data.pop()
        if self.primaryExp is not None:
            if len(exp.arguments)==0:
                prop = self.rootHandler.create_property(self.primaryExp)
                exp.arguments.append(prop)
        self.rootHandler.add_to_root(exp)

        pass

    # Enter a parse tree produced by BOCLParser#SEQUENCE.
    def enterSEQUENCE(self, ctx: BOCLParser.SEQUENCEContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        self.coll_data.append(self.rootHandler.create_sequence())

        pass

    # Exit a parse tree produced by BOCLParser#SEQUENCE.
    def exitSEQUENCE(self, ctx: BOCLParser.SEQUENCEContext):
        op = None
        # print(ctx.getText())
        if len(self.operator) != 0:
            op = self.operator.pop()
            if len(self.coll_data) > 0:
                self.rootHandler.handle_adding_to_root(self.coll_data.pop(), op)

        pass

    # Enter a parse tree produced by BOCLParser#SUBSEQUENCE.
    def enterSUBSEQUENCE(self, ctx: BOCLParser.SUBSEQUENCEContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        self.coll_data.append(self.rootHandler.create_sub_sequence())

        pass

    # Exit a parse tree produced by BOCLParser#SUBSEQUENCE.
    def exitSUBSEQUENCE(self, ctx: BOCLParser.SUBSEQUENCEContext):
        self.rootHandler.add_to_root(self.coll_data.pop())
        pass

    # Enter a parse tree produced by BOCLParser#ALLINSTANCES.
    def enterALLINSTANCES(self, ctx: BOCLParser.ALLINSTANCESContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        if self.userDefined is not None:
            self.rootHandler.add_to_root(self.rootHandler.create_type_exp(self.userDefined))
            self.userDefined = None
        op_call_exp = self.rootHandler.create_operation_call_exp("ALLInstances")
        op_call_exp.referredOperation=self.rootHandler.create_operation_call_exp("ALLInstances")
        self.rootHandler.add_to_root(op_call_exp)
        pass

    # Exit a parse tree produced by BOCLParser#ALLINSTANCES.
    def exitALLINSTANCES(self, ctx: BOCLParser.ALLINSTANCESContext):
        pass

    # Enter a parse tree produced by BOCLParser#ORDEREDSET.
    def enterORDEREDSET(self, ctx: BOCLParser.ORDEREDSETContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        txt = ctx.getText()
        self.coll_data.append(self.rootHandler.create_ordered_set())

        pass

    # Exit a parse tree produced by BOCLParser#ORDEREDSET.
    def exitORDEREDSET(self, ctx: BOCLParser.ORDEREDSETContext):
        op = None
        if len(self.operator) != 0:
            op = self.operator.pop()
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop(), op)
        pass

    # Enter a parse tree produced by BOCLParser#SUBORDEREDSET.
    def enterSUBORDEREDSET(self, ctx: BOCLParser.SUBORDEREDSETContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.coll_data.append(self.rootHandler.create_sub_ordered_set())
        pass

    # Exit a parse tree produced by BOCLParser#SUBORDEREDSET.
    def exitSUBORDEREDSET(self, ctx: BOCLParser.SUBORDEREDSETContext):
        op = None
        if len(self.operator) != 0:
            op = self.operator.pop()
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop(), op)
        pass

    # Enter a parse tree produced by BOCLParser#SET.
    def enterSET(self, ctx: BOCLParser.SETContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.coll_data.append(self.rootHandler.create_set())

        pass

    # Exit a parse tree produced by BOCLParser#SET.
    def exitSET(self, ctx: BOCLParser.SETContext):
        op = None
        if len(self.operator) != 0:
            op = self.operator.pop()
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop(), op)
        pass

    # Enter a parse tree produced by BOCLParser#BAG.
    def enterBAG(self, ctx: BOCLParser.BAGContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        self.coll_data.append(self.rootHandler.create_bag())
        pass

    # Exit a parse tree produced by BOCLParser#BAG.
    def exitBAG(self, ctx: BOCLParser.BAGContext):
        op = None
        if len(self.operator) != 0:
            op = self.operator.pop()
        if len(self.coll_data) > 0:
            self.rootHandler.handle_bag(self.coll_data.pop(), op)

        pass

    # Enter a parse tree produced by BOCLParser#PREPEND.
    def enterPREPEND(self, ctx: BOCLParser.PREPENDContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('PREPEND')
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.coll_data.append(op_call_exp)

        pass

    # Exit a parse tree produced by BOCLParser#PREPEND.
    def exitPREPEND(self, ctx: BOCLParser.PREPENDContext):

        self.rootHandler.add_to_root(self.coll_data.pop())
        pass

    # Enter a parse tree produced by BOCLParser#LAST.
    def enterLAST(self, ctx: BOCLParser.LASTContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('Last')
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.rootHandler.handle_adding_to_root(op_call_exp)

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#LAST.
    def exitLAST(self, ctx: BOCLParser.LASTContext):

        pass

    # Enter a parse tree produced by BOCLParser#APPEND.
    def enterAPPEND(self, ctx: BOCLParser.APPENDContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('APPEND')
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.coll_data.append(op_call_exp)

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#APPEND.
    def exitAPPEND(self, ctx: BOCLParser.APPENDContext):
        self.rootHandler.add_to_root(self.coll_data.pop())
        pass

    # Enter a parse tree produced by BOCLParser#COLLECTION.
    def enterCOLLECTION(self, ctx: BOCLParser.COLLECTIONContext):

        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
                print(ctx.parentCtx.getText())
            # print(ctx.getText())

        if ctx.parentCtx is not None:
            self.rootHandler.verify(ctx.parentCtx.getText().split(ctx.getText())[0])
        self.rootHandler.handle_collection(ctx.getText())
        # print(inspect.stack()[0][3])

    # Exit a parse tree produced by BOCLParser#COLLECTION.
    def exitCOLLECTION(self, ctx: BOCLParser.COLLECTIONContext):
        if self.debug:
            print("exitCOLLECTION")
            if self.debug_print:
                print(ctx.getText())

        self.rootHandler.pop()

        pass

    # Enter a parse tree produced by BOCLParser#CollectionExpressionVariable.
    def enterCollectionExpressionVariable(self, ctx: BOCLParser.CollectionExpressionVariableContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
                print(ctx.parentCtx.getText())
        if ctx.parentCtx is not None:
            self.rootHandler.verify(ctx.parentCtx.getText().split(ctx.getText())[0])
        self.rootHandler.handle_collection(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#CollectionExpressionVariable.
    def exitCollectionExpressionVariable(self, ctx: BOCLParser.CollectionExpressionVariableContext):
        # print("exitCollectionExpressionVariable")
        # self.rootHandler.pop()
        pass

    # Enter a parse tree produced by BOCLParser#SYMMETRICDIFFERENCE.
    def enterSYMMETRICDIFFERENCE(self, ctx: BOCLParser.SYMMETRICDIFFERENCEContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('SYMMETRICDIFFERENCE')
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.coll_data.append(op_call_exp)
        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#SYMMETRICDIFFERENCE.
    def exitSYMMETRICDIFFERENCE(self, ctx: BOCLParser.SYMMETRICDIFFERENCEContext):
        self.rootHandler.add_to_root(self.coll_data.pop())

    # Enter a parse tree produced by BOCLParser#FIRST.
    def enterFIRST(self, ctx: BOCLParser.FIRSTContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('First')
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.rootHandler.handle_adding_to_root(op_call_exp)
        # print(inspect.stack()[0][3])

    # Exit a parse tree produced by BOCLParser#FIRST.
    def exitFIRST(self, ctx: BOCLParser.FIRSTContext):
        pass

    # Enter a parse tree produced by BOCLParser#DERIVE.
    def enterDERIVE(self, ctx: BOCLParser.DERIVEContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#DERIVE.
    def exitDERIVE(self, ctx: BOCLParser.DERIVEContext):
        pass

    # Enter a parse tree produced by BOCLParser#UNION.
    def enterUNION(self, ctx: BOCLParser.UNIONContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        op_call_exp = self.rootHandler.create_operation_call_exp('UNION')
        if len(self.coll_data) > 0:
            self.rootHandler.handle_adding_to_root(self.coll_data.pop())
        self.coll_data.append(op_call_exp)

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#UNION.
    def exitUNION(self, ctx: BOCLParser.UNIONContext):
        self.rootHandler.add_to_root(self.coll_data.pop())

        pass

    # Enter a parse tree produced by BOCLParser#defExp.
    def enterDefExp(self, ctx: BOCLParser.DefExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#defExp.
    def exitDefExp(self, ctx: BOCLParser.DefExpContext):
        pass

    # Enter a parse tree produced by BOCLParser#defIDAssignmentexpression.
    def enterDefIDAssignmentexpression(self, ctx: BOCLParser.DefIDAssignmentexpressionContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#defIDAssignmentexpression.
    def exitDefIDAssignmentexpression(self, ctx: BOCLParser.DefIDAssignmentexpressionContext):
        pass

    # Enter a parse tree produced by BOCLParser#PrimaryExp.
    def enterPrimaryExp(self, ctx: BOCLParser.PrimaryExpContext):
        if self.unary:
            self.sign.append(self.unary.split(ctx.getText())[0])
            print(self.sign)
            if self.unary == self.sign[-1] + ctx.getText():
                self.rootHandler.handle_single_variable(ctx.getText(), self.sign.pop())
                #
        self.primaryExp = ctx.getText()
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())


        # print(inspect.stack()[0][3])
        # print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#PrimaryExp.
    def exitPrimaryExp(self, ctx: BOCLParser.PrimaryExpContext):
        # #print("exitPrimaryExp")
        # #print(ctx.getText())
        # input()
        # if len(self.operator) !=0 and self.operator[-1] in ctx.getText():
        #
        #     self.rootHandler.handlePrimaryExp(ctx.getText(),self.operator[-1])
        pass

    # Enter a parse tree produced by BOCLParser#funcCall.
    def enterFuncCall(self, ctx: BOCLParser.FuncCallContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#funcCall.
    def exitFuncCall(self, ctx: BOCLParser.FuncCallContext):
        pass

    # Enter a parse tree produced by BOCLParser#op.
    def enterOp(self, ctx: BOCLParser.OpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        # self.operator.append(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#op.
    def exitOp(self, ctx: BOCLParser.OpContext):
        pass

    # Enter a parse tree produced by BOCLParser#arrowexp.
    def enterArrowexp(self, ctx: BOCLParser.ArrowexpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        # print(inspect.stack()[0][3])
        pass

    # Exit a parse tree produced by BOCLParser#arrowexp.
    def exitArrowexp(self, ctx: BOCLParser.ArrowexpContext):
        pass

    # Enter a parse tree produced by BOCLParser#number.
    def enterNumber(self, ctx: BOCLParser.NumberContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        item = self.rootHandler.get_factory().create_collection_item("item", ctx.getText())
        self.initItems.append(item)
        if len(self.coll_data) > 0:
            self.coll_data[-1].add(item)

        pass

    # Exit a parse tree produced by BOCLParser#number.
    def exitNumber(self, ctx: BOCLParser.NumberContext):
        pass

    # Enter a parse tree produced by BOCLParser#PredefinedfunctionCall.
    def enterPredefinedfunctionCall(self, ctx: BOCLParser.PredefinedfunctionCallContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#PredefinedfunctionCall.
    def exitPredefinedfunctionCall(self, ctx: BOCLParser.PredefinedfunctionCallContext):
        pass

    # Enter a parse tree produced by BOCLParser#ID.
    def enterID(self, ctx: BOCLParser.IDContext):
        # print(inspect.stack()[0][3])
        # print(ctx.getText())
        txt = ctx.getText()

        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        if len(self.coll_data) != 0:
            item = self.rootHandler.get_factory().create_collection_item("item", ctx.getText())
            self.coll_data[-1].add(item)
        else:
            if self.rootHandler.get_root() is not None:
                self.rootHandler.handle_ID(ctx.getText())
            if len(self.rootHandler.all)>0:
                self.rootHandler.handle_ID(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#ID.
    def exitID(self, ctx: BOCLParser.IDContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        item = self.rootHandler.get_factory().create_collection_item("item", ctx.getText())
        self.initItems.append(item)

    # Enter a parse tree produced by BOCLParser#SingleQuoteExp.
    def enterSingleQuoteExp(self, ctx: BOCLParser.SingleQuoteExpContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#SingleQuoteExp.
    def exitSingleQuoteExp(self, ctx: BOCLParser.SingleQuoteExpContext):
        pass

    # Enter a parse tree produced by BOCLParser#doubleDots.
    def enterDoubleDots(self, ctx: BOCLParser.DoubleDotsContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#doubleDots.
    def exitDoubleDots(self, ctx: BOCLParser.DoubleDotsContext):
        pass

    # Enter a parse tree produced by BOCLParser#binaryExpression.
    def enterBinaryExpression(self, ctx: BOCLParser.BinaryExpressionContext):

        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#binaryExpression.
    def exitBinaryExpression(self, ctx: BOCLParser.BinaryExpressionContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        and_or = self.binary.split(ctx.getText())[0]
        # print(and_or)
        beforeSign = None
        if len(self.sign) != 0:
            beforeSign = self.sign.pop()
        inBetween_op = None
        if "and" or "or" in and_or:
            inBetween_op = and_or
        if len(self.operator) != 0:
            text_to_process = ctx.getText()
            if "()" in text_to_process:
                text_to_process = ctx.parentCtx.parentCtx.getText()
            self.rootHandler.handle_binary_expression(text_to_process, self.operator.pop(), inBetween_op, beforeSign)
            # self.operator.pop()

        pass

    # Enter a parse tree produced by BOCLParser#unaryExpression.
    def enterUnaryExpression(self, ctx: BOCLParser.UnaryExpressionContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#unaryExpression.
    def exitUnaryExpression(self, ctx: BOCLParser.UnaryExpressionContext):
        pass

    # Enter a parse tree produced by BOCLParser#operator.
    def enterOperator(self, ctx: BOCLParser.OperatorContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        self.operator.append(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#operator.
    def exitOperator(self, ctx: BOCLParser.OperatorContext):
        pass

    # Enter a parse tree produced by BOCLParser#numberORUserDefined.
    def enterNumberORUserDefined(self, ctx: BOCLParser.NumberORUserDefinedContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        self.userDefined = ctx.getText()

        pass

    # Exit a parse tree produced by BOCLParser#numberORUserDefined.
    def exitNumberORUserDefined(self, ctx: BOCLParser.NumberORUserDefinedContext):
        pass

    # Enter a parse tree produced by BOCLParser#primaryExpression.
    def enterPrimaryExpression(self, ctx: BOCLParser.PrimaryExpressionContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        # print(ctx.getText())
        # print(ctx.getText())
        if not ctx.getText() == "self":
            if len(self.types_of) != 0:
                self.rootHandler.handle_property(ctx.getText())

        self.userDefined = ctx.getText()
        pass

    # Exit a parse tree produced by BOCLParser#primaryExpression.
    def exitPrimaryExpression(self, ctx: BOCLParser.PrimaryExpressionContext):
        # if len(self.operator) != 0:
        #     self.rootHandler.handlePrimaryExp(ctx.getText(),self.operator[-1])
        #     self.operator.pop()
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Enter a parse tree produced by BOCLParser#literal.
    def enterLiteral(self, ctx: BOCLParser.LiteralContext):
        # print(inspect.stack()[0][3])
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        pass

    # Exit a parse tree produced by BOCLParser#literal.
    def exitLiteral(self, ctx: BOCLParser.LiteralContext):
        pass

    def print(self):
        self.rootHandler.print()

    def enterIfExp(self, ctx: BOCLParser.IfExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        self.all_if_else.append(self.rootHandler.create_if_else_exp('if', 'IfExpression'))

        pass

    # Exit a parse tree produced by BOCLParser#ifExp.
    def exitIfExp(self, ctx: BOCLParser.IfExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Enter a parse tree produced by BOCLParser#thenExp.
    def enterThenExp(self, ctx: BOCLParser.ThenExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        top = self.rootHandler.getTop()
        self.all_if_else[-1].ifCondition = top

    # Exit a parse tree produced by BOCLParser#thenExp.
    def exitThenExp(self, ctx: BOCLParser.ThenExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Enter a parse tree produced by BOCLParser#elseExp.
    def enterElseExp(self, ctx: BOCLParser.ElseExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        self.all_if_else[-1].thenExpression = self.rootHandler.getTop()
        pass

    # Exit a parse tree produced by BOCLParser#elseExp.
    def exitElseExp(self, ctx: BOCLParser.ElseExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Enter a parse tree produced by BOCLParser#endIfExp.
    def enterEndIfExp(self, ctx: BOCLParser.EndIfExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        self.all_if_else[-1].elseCondition = self.rootHandler.getTop(True)
        self.rootHandler.add_to_root(self.all_if_else.pop())

    # Exit a parse tree produced by BOCLParser#endIfExp.
    def exitEndIfExp(self, ctx: BOCLParser.EndIfExpContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    def enterDoubleCOLONs(self, ctx: BOCLParser.DoubleCOLONsContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())

        self.userDefined = ctx.getText().split('::')[0]
        pass

    # Exit a parse tree produced by BOCLParser#doubleCOLONs.
    def exitDoubleCOLONs(self, ctx: BOCLParser.DoubleCOLONsContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    def enterEndExpression(self, ctx: BOCLParser.EndExpressionContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        self.rootHandler.pop()
        if len(ctx.getText())>2:
            if ctx.getText()[0:3] == "and" or ctx.getText()[0:2] == "or":
                self.rootHandler.handle_and_with_function_call(ctx.getText())
        # elif ctx.getText()[0:1] in ['+','-','*','/','=']:
        #     self.rootHandler.create_infinix_op(ctx.getText()[0:1])

        pass
# Exit a parse tree produced by BOCLParser#endExpression.
    def exitEndExpression(self, ctx: BOCLParser.EndExpressionContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass
    def enterBinaryFunctionCall(self, ctx:BOCLParser.BinaryFunctionCallContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#binaryFunctionCall.
    def exitBinaryFunctionCall(self, ctx:BOCLParser.BinaryFunctionCallContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        operator = self.operator.pop()
        num = ctx.getText().replace(operator,"")
        self.rootHandler.handleBinaryFunc(operator,num)
# Enter a parse tree produced by BOCLParser#dateLiteral.
    def enterDateLiteral(self, ctx:BOCLParser.DateLiteralContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass

    # Exit a parse tree produced by BOCLParser#dateLiteral.
    def exitDateLiteral(self, ctx:BOCLParser.DateLiteralContext):
        if self.debug:
            print(inspect.stack()[0][3])
            if self.debug_print:
                print(ctx.getText())
        pass
del BOCLParser


"""
BUML code for the NN available in this paper: 
'LSTM and GRU neural network performance comparison study'
https://github.com/g9g99g9g/basic/blob/master/ANN/LSTM_vs_GRU/yelp_reviews_emotion_prediction.py
"""

from besser.BUML.metamodel.nn import NN, LinearLayer, DropoutLayer, \
   EmbeddingLayer, LSTMLayer
from besser.generators.nn.pytorch.pytorch_code_generator import (
    PytorchGenerator
)
from besser.generators.nn.tf.tf_code_generator import TFGenerator


lstm_model: NN = NN(name="lstm")
lstm_model.add_layer(EmbeddingLayer(name="l1", actv_func=None,
                                    num_embeddings=10000, embedding_dim=326))
lstm_model.add_layer(LSTMLayer(name="l2", actv_func=None, return_type="full",
                               input_size=326, hidden_size=40,
                               bidirectional=True, dropout=0.5,
                               batch_first=True))
lstm_model.add_layer(DropoutLayer(name="l3", rate=0.2))
lstm_model.add_layer(LSTMLayer(name="l4", actv_func=None, return_type="last",
                               input_size=2*40, hidden_size=40, dropout=0.2,
                               batch_first=True))
lstm_model.add_layer(LinearLayer(name="l5", actv_func="relu", in_features=40,
                                 out_features=40))
lstm_model.add_layer(LinearLayer(name="l6", actv_func="softmax",
                                 in_features=40, out_features=2))


pytorch_model = PytorchGenerator(model=lstm_model, output_dir="output/lstm")
pytorch_model.generate()
tf_model = TFGenerator(model=lstm_model, output_dir="output/lstm")
tf_model.generate()



"""
This module provides helper functions to convert BUML code
to PyTorch or TensorFlow code.
"""

import os
import random

from PIL import Image
import numpy as np
from torch import nn


def get_previous_out_var(modules_details, prev_module):
    """
    It retrieves the output variable of the previous module in order to
    use it as the input variable of the current module.
    """
    if isinstance(modules_details[prev_module], dict):
        return modules_details[prev_module]["in_out_variable"]
    else:
        return modules_details[prev_module][1]

def get_input_var(layer, modules_details, prev_out_var):
    """
    It determines the input variable of the current layer. It is either
    the output variable of the module in `name_module_input` attribute 
    (if it is given), or simply the output of the previous module given
    by `get_previous_out_variable` function.
    """
    modules_names = list(modules_details.keys())
    lyr_input = layer.name_module_input
    if lyr_input is not None:
        if f"{lyr_input}_layer" in modules_names:
            return modules_details[f"{lyr_input}_layer"][1]
        if f"{lyr_input}_nn" in modules_names:
            return  modules_details[f"{lyr_input}_nn"]["in_out_variable"]
        #module.name_module_input+"_op" in my_keys
        return modules_details[f"{lyr_input}_op"][1]
    return prev_out_var

def add_in_out_var_to_subnn(modules_details):
    """
    It sets the in_out_variable of subnns, which refers to the input
    and output variable of the subnn.
    """
    last_module = list(modules_details.keys())[-1]
    if len(modules_details) == 1:
        in_out_var = "x"
    else:
        prev_module_details = list(modules_details.values())[-2]
        if isinstance(prev_module_details, dict):
            in_out_var = prev_module_details["in_out_variable"]
        else:
            # get the output variable of the layer or tns before the sub_nn
            in_out_var = prev_module_details[1]
    modules_details[last_module]["in_out_variable"] = in_out_var
    return modules_details

def get_layers_output_for_tensorops(layers_names, modules_details):
    """
    It retrieves the output variables of the layers in `layers_name`
    list to use them as input of the tensorop.
    """
    my_keys = list(modules_details.keys())
    out_vars = []
    for layer_name in layers_names:
        if layer_name+"_layer" in my_keys:
            out_vars.append(modules_details[layer_name + "_layer"][1])
        else:
            out_vars.append(modules_details[layer_name + "_op"][1])
    return out_vars

def initialize_tensorop_var(tensorop):
    """
    It sets the output variable of the tensorop in the case it is the
    first module in the neural network.
    """
    if tensorop.input_reused is True:
        out_var = "x_1"
    else:
        out_var = "x"
    return out_var


def get_out_var_input_reused(prev_out_var):
    """
    It sets the output variable of the module in the case the output
    of the previous module is reused (therefore, they need to be 
    different).
    """
    if prev_out_var == "x":
        out_var = "x_1"
    else:
        out_var = f"x_{int(prev_out_var.split('_')[-1])+1}"
    return out_var


def get_layer_vars(layer, prev_out_var, modules_details):
    """
    It sets the input and output variables of the layer.
    """
    out_var_actv, in_var_actv = None, None
    if layer.input_reused:
        out_var_layer = get_out_var_input_reused(prev_out_var)
    else:
        out_var_layer = prev_out_var
    in_var_layer = get_input_var(layer, modules_details,
                                           prev_out_var)
    if layer.actv_func is not None:
        out_var_actv, in_var_actv = out_var_layer, out_var_layer
    return out_var_layer, in_var_layer, out_var_actv, in_var_actv

def initialize_layer_vars(layer):
    """
    It sets the input and output variables of layer (and activation 
    function for PyTorch) in the case it is the first module in 
    the neural network.
    """
    out_var_actv, in_var_actv = None, None
    if layer.input_reused is True:
        out_var_layer, in_var_layer = "x_1", "x"
        if layer.actv_func is not None:
            out_var_actv, in_var_actv = "x_1", "x_1"
    else:
        out_var_layer, in_var_layer = "x", "x"
        if layer.actv_func is not None:
            out_var_actv, in_var_actv = "x", "x"
    return out_var_layer, in_var_layer, out_var_actv, in_var_actv



def get_layer_syntax(setup_layer_cls, layer, modules_details,
                     in_layer, actv_func_synt):
    """
    It retrieves the syntax of the layer (and the activation 
    function in the case of PyTorch) from the ´setup_layer_cls´ 
    class.
    """
    setup = setup_layer_cls(layer, modules_details)
    parent_class = layer.__class__.mro()[1].__name__
    if (parent_class == "ConvolutionalLayer" or parent_class == "CNN"):
        layer_synt, modules_details = setup.setup_cnn()
    elif parent_class == "RNN":
        layer_synt, modules_details = setup.setup_rnn()
    elif parent_class == "GeneralLayer":
        layer_synt = setup.setup_general_layer()
    else: #(parent_class == "LayerModifier" or
           #parent_class == "NormalizationLayer")
        layer_synt = setup.setup_layer_modifier()

    if actv_func_synt:
        actv_func_synt = setup.setup_actv_func()

    return layer_synt, actv_func_synt, modules_details, setup

def handle_layer(layer, setup_layer, modules_details, actv_func_syntax=False,
                 is_seq=False, channel_last=True, is_subnn=False):
    """
    It populates the `modules_details` dictionary with layer's 
    information: Its syntax, input and output variables, and the 
    layer class.
    In the case of PyTorch, the activation function is treated as
    a layer.
    """

    if len(modules_details) == 0:
        out_layer, in_layer, out_actv, in_actv = initialize_layer_vars(layer)
    else:
        prev_module = list(modules_details.keys())[-1]
        prev_out_var = get_previous_out_var(modules_details, prev_module)
        out_layer, in_layer, out_actv, in_actv = get_layer_vars(
            layer, prev_out_var, modules_details)

    layer_synt, actv_func_syntax, modules_details, setup = get_layer_syntax(
        setup_layer, layer, modules_details, in_layer, actv_func_syntax)

    if setup.permute_in and channel_last:
        dim = setup.dim
        setup.add_permute(layer.name, dim, in_layer, permute_in=True,
                          sequential=is_seq, is_subnn=is_subnn)

    modules_details[layer.name + "_layer"] = [layer_synt, out_layer,
                                              in_layer, layer]
    if actv_func_syntax:
        modules_details[layer.name + "_activ"] = [actv_func_syntax, out_actv,
                                                  in_actv]
    if setup.permute_out and channel_last:
        dim = setup.dim
        setup.add_permute(layer.name, dim, out_layer, permute_in = False,
                          sequential=is_seq, is_subnn=is_subnn)
    return modules_details


def get_tensorop_params(tensorop, modules_details):
    """
    It retrieves tensorops parameters that are used by 
    `get_tensorop_syntax` function defined in PyTorch and 
    TensorFlow `utils.py` files
    """
    if len(list(modules_details.keys())) == 0:
        prev_out_var = "x"
    else:
        prev_module = list(modules_details.keys())[-1]
        prev_out_var = get_previous_out_var(modules_details, prev_module)
    tns_type = tensorop.tns_type
    if tns_type == "reshape":
        params = ', '.join([str(i) for i in tensorop.reshape_dim])
    elif tns_type == "concatenate":
        tensors = get_layers_output_for_tensorops(tensorop.layers_of_tensors,
                                                  modules_details)
        params = ', '.join(tensors)
    elif tns_type == "transpose":
        params = ", ".join([str(i) for i in tensorop.transpose_dim])
    elif tns_type == "permute":
        params = ", ".join([str(i) for i in tensorop.permute_dim])
    else:
        tensors = tensorop.layers_of_tensors
        if isinstance(tensors[0], str):
            tensors = get_layers_output_for_tensorops(tensorop,
                                                      modules_details)

        params = ', '.join([str(i) for i in tensors])
    return prev_out_var, params


def get_tensorop_out_var(tensorop, prev_out_var):
    """
    It sets the output variable of tensorop.
    """
    if tensorop.input_reused is True:
        out_var  = get_out_var_input_reused(prev_out_var)
    else:
        out_var = prev_out_var
    return out_var

def handle_tensorop(tensorop, modules_details,
                    get_tensorop_syntax, out_var=None):
    """
    It populates the `modules_details` dictionary with tensorop's 
    information: Its syntax and output variable.
    """
    ts_op_synt = get_tensorop_syntax(tensorop, modules_details, out_var)
    if out_var is None:
        if len(modules_details) == 0:
            out_var  = initialize_tensorop_var(tensorop)
        else:
            prev_module = list(modules_details.keys())[-1]
            prev_out_var = get_previous_out_var(modules_details, prev_module)
            out_var = get_tensorop_out_var(tensorop, prev_out_var)

    modules_details[tensorop.name + "_op"] = [ts_op_synt, out_var]
    return modules_details


def preprocess_image(image_path, target_size):
    """
    It resizes and returns the images as np arrays.
    """
    image = Image.open(image_path)
    image = image.resize(target_size)
    np_image = np.array(image.convert('RGB'), dtype=np.float32)
    return np_image



def compute_mean_std(image_dir, num_samples=100, target_size=(256, 256)):
    """
    It computes the mean and standard deviation of images and checks
    whether scaling is needed.
    """
    image_files = [os.path.join(root, file)
                   for root, _, files in os.walk(image_dir)
                   for file in files]
    sampled_files = random.sample(image_files,
                                  min(num_samples, len(image_files)))
    all_pixels = []
    for file in sampled_files:
        np_image = preprocess_image(file, target_size)
        all_pixels.append(np_image.reshape(-1, 3))

    all_pixels = np.concatenate(all_pixels)

    # Rescale if necessary
    if all_pixels.max() >= 1:
        scale = True
        all_pixels /= 255.0
    else:
        scale=False

    return (scale, np.mean(all_pixels, axis=0).tolist(),
            np.std(all_pixels, axis=0).tolist())


def format_value(elem):
    """
    It formats BUML list of int. If it contains one element, it is
    returned as `int`. Otherwise, it converts the list to a tuple.
    """
    if len(elem) == 1:
        return elem[0]
    return tuple(elem)


class Permute(nn.Module):
    """A custom permute module for the sequential architecture"""
    def __init__(self, dims):
        super().__init__()
        self.dims = dims

    def forward(self, x):
        x = x.permute(self.dims)
        return x



from antlr4 import CommonTokenStream, FileStream, ParseTreeWalker
from .deploymentLexer import deploymentLexer
from .deploymentParser import deploymentParser
from .depl_to_buml_listener import Deployment_BUML_Listener
from besser.BUML.metamodel.deployment import DeploymentModel
import os

def buml_deployment_model(deployment_textfile: str, buml_model_file_name: str = "deployment_buml_model", output_dir: str = "buml"):
    """
    Args:
        deployment_textfile (str): The path to the file containing the deployment architecture textual definition.
        buml_model_file_name (str, optional): the name of the file produced with the base code to build the B-UML model.
        output_dir (str, optional): directory where the intermediate file will be stored. Defaults to "buml".

    Returns:
        BUML_model (DeploymentModel): the B-UML model object.
    """
    lexer = deploymentLexer(FileStream(deployment_textfile))
    parser = deploymentParser(CommonTokenStream(lexer))
    parse_tree = parser.architecture()

    # file creation with proper path handling
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    output_file = os.path.join(output_dir, f"{buml_model_file_name}.py") if output_dir else f"{buml_model_file_name}.py"

    with open(output_file, "w+") as output:
        listen = Deployment_BUML_Listener(output)
        walker = ParseTreeWalker()
        walker.walk(listen, parse_tree)

    # model creation
    namespace = {}
    with open(output_file, 'r') as model_code:
        code = model_code.read()
        exec(code, namespace)
    
    BUML_model: DeploymentModel = namespace.get('deployment_model')
    return BUML_model